import copy
import unittest

from hamcrest import assert_that, equal_to
from mock import patch, call, MagicMock

from zmote.discoverer import Discoverer

_UUID = 'CI00a1b2c3'

_TEST_RESPONSE = 'AMXB<-UUID={0}><-Type=ZMT2><-Make=zmote.io><-Model=ZV-2><-Revision=2.1.4><-Config-URL=http://192.168.1.12>'.format(_UUID)

_TEST_RESPONSE_PARSED = {
    'UUID': 'CI00a1b2c3',
    'Make': 'zmote.io',
    'Config-URL': 'http://192.168.1.12',
    'Model': 'ZV-2',
    'Type': 'ZMT2',
    'Revision': '2.1.4'
}

_TEST_RESPONSE_PARSED_WITH_IP = {
    'IP': '192.168.1.12',
    'UUID': 'CI00a1b2c3',
    'Make': 'zmote.io',
    'Config-URL': 'http://192.168.1.12',
    'Model': 'ZV-2',
    'Type': 'ZMT2',
    'Revision': '2.1.4'
}


class DiscovererTest(unittest.TestCase):
    @patch('zmote.discoverer.socket')
    def setUp(self, socket):
        socket.inet_aton.return_value = '\xef\xff\xc0\x89'
        socket.AF_INET = 1
        socket.SOCK_DGRAM = 2
        socket.IPPROTO_UDP = 3
        socket.SOL_SOCKET = 4
        socket.SO_REUSEADDR = 5
        socket.IPPROTO_IP = 6
        socket.INADDR_ANY = 7
        socket.IP_ADD_MEMBERSHIP = 8

        self._subject = Discoverer()

        assert_that(
            socket.mock_calls,
            equal_to([
                call.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP),
                call.socket().setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
                call.inet_aton('239.255.250.250'),
                call.socket().setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP,
                                         '\xef\xff\xc0\x89\x00\x00\x00\x00\x07\x00\x00\x00\x00\x00\x00\x00')]
            )
        )

        self._subject._sock.mock_calls = []

    def test_bind(self):
        self._subject.bind()

        assert_that(
            self._subject._sock.mock_calls,
            equal_to([
                call.bind(('239.255.250.250', 9131))
            ])
        )

    @patch('zmote.discoverer.socket')
    def test_send(self, socket):
        socket.AF_INET = 1
        socket.SOCK_DGRAM = 2
        socket.IPPROTO_UDP = 3

        self._subject.send()

        assert_that(
            socket.mock_calls,
            equal_to([
                call.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP),
                call.socket().sendto('SENDAMXB', ('239.255.250.250', 9130))
            ])
        )

    def test_receive(self):
        self._subject._sock.recv.return_value = _TEST_RESPONSE

        assert_that(
            self._subject.receive(),
            equal_to(_TEST_RESPONSE)
        )

    def test_parse(self):
        assert_that(
            self._subject.parse(_TEST_RESPONSE),
            equal_to(_TEST_RESPONSE_PARSED)
        )

    def test_discover_limit(self):
        self._subject.receive = MagicMock()
        self._subject.parse = MagicMock()
        self._subject.parse.return_value = copy.deepcopy(_TEST_RESPONSE_PARSED)

        assert_that(
            self._subject.discover(unique_zmote_limit=1),
            equal_to({
                _UUID: _TEST_RESPONSE_PARSED_WITH_IP,
            })
        )

    def test_discover_uuid(self):
        self._subject.receive = MagicMock()
        self._subject.parse = MagicMock()
        self._subject.parse.return_value = copy.deepcopy(_TEST_RESPONSE_PARSED)

        assert_that(
            self._subject.discover(uuid_to_look_for='CI00a1b2c3'),
            equal_to({
                _UUID: _TEST_RESPONSE_PARSED_WITH_IP,
            })
        )



import inspect
import socket
import struct
import time
from logging import getLogger

socket.setdefaulttimeout(30)

_GROUP = '239.255.250.250'
_RECEIVE_PORT = 9131
_SEND_PORT = 9130


class Discoverer(object):
    def __init__(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        mreq = struct.pack("4sl", socket.inet_aton(_GROUP), socket.INADDR_ANY)
        self._sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

        self._logger = getLogger(self.__class__.__name__)
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name
        ))

    def bind(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name,
        ))

        self._sock.bind((_GROUP, _RECEIVE_PORT))

    def send(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name,
        ))

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.sendto('SENDAMXB', (_GROUP, _SEND_PORT))

    def receive(self):
        data = self._sock.recv(1024)

        self._logger.debug('{0}(); data={1}'.format(
            inspect.currentframe().f_code.co_name, repr(data)
        ))

        return data

    def parse(self, data):
        self._logger.debug('{0}({1})'.format(
            inspect.currentframe().f_code.co_name, repr(data)
        ))

        if not data.startswith('AMXB') or not all([x in data for x in ['UUID', 'Type', 'Make', 'Model', 'Revision', 'Config-URL']]):
            exception = ValueError(
                'cannot parse data {0}; does not appear to be in correct format'.format(
                    repr(data)
                )
            )

            self._logger.error('{0}({1}); exception={1}'.format(
                inspect.currentframe().f_code.co_name, repr(data)
            ))

            raise exception

        data = dict([x[0:-1].split('=') for x in data.split('<-')[1:]])

        self._logger.debug('{0}({1}); data={1}'.format(
            inspect.currentframe().f_code.co_name, repr(data)
        ))

        return data

    def discover(self, unique_zmote_limit=None, uuid_to_look_for=None):
        if unique_zmote_limit is not None and uuid_to_look_for is not None:
            raise ValueError('must specify only one (or neither) of unique_zmote_limit or uuid_to_look_for')

        zmotes_by_uuid = {}

        self._logger.debug('{0}({1})'.format(
            inspect.currentframe().f_code.co_name, unique_zmote_limit
        ))

        def keep_waiting():
            if unique_zmote_limit is not None:
                if len(zmotes_by_uuid) >= unique_zmote_limit:
                    return False
            elif uuid_to_look_for in zmotes_by_uuid:
                return False

            return True

        while keep_waiting():
            try:
                data = self.receive()
            except socket.timeout:
                break

            parsed_data = self.parse(data)

            ip = parsed_data.get('Config-URL').split('//')[-1].strip('/')
            parsed_data.update({
                'IP': ip,
            })

            zmotes_by_uuid.update({
                parsed_data['UUID']: parsed_data,
            })

        self._logger.debug('{0}({1}); zmotes_by_uuid={2}'.format(
            inspect.currentframe().f_code.co_name, unique_zmote_limit, zmotes_by_uuid
        ))

        return zmotes_by_uuid


def passive_discover_zmotes(unique_zmote_limit, uuid_to_look_for):
    d = Discoverer()
    d.bind()
    return d.discover(
        unique_zmote_limit=unique_zmote_limit,
        uuid_to_look_for=uuid_to_look_for,
    )


def active_discover_zmotes(unique_zmote_count, uuid_to_look_for):
    d = Discoverer()
    d.bind()
    for i in range(0, 5):
        d.send()
        time.sleep(0.1)
    return d.discover(
        unique_zmote_limit=unique_zmote_count,
        uuid_to_look_for=uuid_to_look_for,
    )


if __name__ == '__main__':
    import argparse
    import pprint

    import logging

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    )

    logger = logging.getLogger(Discoverer.__name__)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    parser = argparse.ArgumentParser(
        description='Discover zmote.io devices on the local network- default behaviour with no arguments is to time out after 30 seconds',
        epilog='See http://www.zmote.io/apis for more detail'
    )

    parser.add_argument(
        '-l',
        '--unique-zmote-limit',
        type=int,
        default=None,
        help='number of unique devices to wait for',
    )

    parser.add_argument(
        '-u',
        '--uuid-to-look-for',
        type=str,
        default=None,
        help='UUID of a device to wait for',
    )

    parser.add_argument(
        '-a',
        '--active',
        action='store_true',
        required=False,
        help='send probe first to discover devices faster (default disabled)',
    )

    args = parser.parse_args()

    if args.unique_zmote_limit is not None and args.uuid_to_look_for is not None:
        parser.error('must specify only one (or neither) of --unique-zmote-limit or --uuid-to-look-for')

    if args.active:
        zmotes = active_discover_zmotes(args.unique_zmote_limit, args.uuid_to_look_for)
    else:
        zmotes = passive_discover_zmotes(args.unique_zmote_limit, args.uuid_to_look_for)

    print ''
    pprint.pprint(zmotes)



import inspect
import socket
from logging import getLogger

from requests import Session

socket.setdefaulttimeout(5)


class HTTPTransport(object):
    def __init__(self, ip):
        self._ip = ip

        self._session = None
        self._uuid = None

        self._logger = getLogger(self.__class__.__name__)
        self._logger.debug('{0}(); ip={1}'.format(
            inspect.currentframe().f_code.co_name, repr(ip)
        ))

    def get_uuid(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name,
        ))

        uuid = self._session.get(
            'http://{0}/uuid'.format(self._ip),
            timeout=5,
        ).text.split(',')[-1].strip()

        self._logger.debug('{0}(); uuid={1}'.format(
            inspect.currentframe().f_code.co_name, repr(uuid)
        ))

        return uuid

    def connect(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name,
        ))

        self._session = Session()
        self._uuid = self.get_uuid()

        self._logger.debug('{0}(); session={1}, uuid={2}'.format(
            inspect.currentframe().f_code.co_name, self._session, repr(self._uuid)
        ))

    def call(self, data):
        self._logger.debug('{0}({1})'.format(
            inspect.currentframe().f_code.co_name, repr(data),
        ))

        output = self._session.post(
            url='http://{0}/v2/{1}'.format(
                self._ip, self._uuid,
            ),
            data=data,
            timeout=5,
        ).text

        self._logger.debug('{0}({1}); output={2}'.format(
            inspect.currentframe().f_code.co_name, repr(data), repr(output)
        ))

        return output

    def disconnect(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name,
        ))

        self._session = None


class TCPTransport(object):
    def __init__(self, ip):
        self._ip = ip

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self._logger = getLogger(self.__class__.__name__)
        self._logger.debug('{0}(); ip={1}'.format(
            inspect.currentframe().f_code.co_name, repr(ip)
        ))

    def connect(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name,
        ))

        self._sock.connect((self._ip, 4998))

    def call(self, data):
        self._logger.debug('{0}({1})'.format(
            inspect.currentframe().f_code.co_name, repr(data),
        ))

        self._sock.send(data)

        buf = self._sock.recv(1024)
        if 'IR Learner Enabled' in buf:
            buf += self._sock.recv(1024)

        self._logger.debug('{0}({1}); buf={2}'.format(
            inspect.currentframe().f_code.co_name, repr(data), repr(buf)
        ))

        return buf

    def disconnect(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name,
        ))

        self._sock.close()


class Connector(object):
    def __init__(self, transport):
        self._transport = transport

        self._logger = getLogger(self.__class__.__name__)
        self._logger.debug('{0}(); transport={1}'.format(
            inspect.currentframe().f_code.co_name, transport
        ))

    def connect(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name
        ))

        self._transport.connect()

    def send(self, data):
        self._logger.debug('{0}({1})'.format(
            inspect.currentframe().f_code.co_name, repr(data)
        ))

        data = data.split('sendir,')[-1]

        output = self._transport.call('sendir,{0}'.format(data))

        self._logger.debug('{0}({1}); output={2}'.format(
            inspect.currentframe().f_code.co_name, repr(data), repr(output)
        ))

        return output

    def learn(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name
        ))

        data = self._transport.call('get_IRL')

        self._logger.debug('{0}(); data={1}'.format(
            inspect.currentframe().f_code.co_name, repr(data)
        ))

        return data.split('sendir,')[-1]

    def disconnect(self):
        self._logger.debug('{0}()'.format(
            inspect.currentframe().f_code.co_name
        ))

        self._transport.disconnect()


if __name__ == '__main__':
    import argparse

    import logging

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    )

    for cls in [HTTPTransport, TCPTransport, Connector]:
        logger = logging.getLogger(cls.__name__)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    parser = argparse.ArgumentParser(
        description='Make calls against a zmote.io device via HTTP or TCP',
        epilog='See http://www.zmote.io/apis for more detail',
    )

    parser.add_argument(
        '-t',
        '--transport',
        required=False,
        choices=['http', 'tcp'],
        default='http',
        help='transport to use for communication with the device (default HTTP)',
    )

    parser.add_argument(
        '-d',
        '--device-ip-or-hostname',
        type=str,
        required=True,
        help='IP or hostname of the device',
    )

    parser.add_argument(
        '-c',
        '--call-type',
        choices=['learn', 'send'],
        required=True,
        help='type of call to make to the device'
    )

    parser.add_argument(
        '-p',
        '--payload',
        type=str,
        default=None,
        help='payload to send (applicable for send call-type only)',
    )

    args = parser.parse_args()

    transport = None
    if args.transport == 'http':
        transport = HTTPTransport(
            ip=args.device_ip_or_hostname,
        )
    elif args.transport == 'tcp':
        transport = TCPTransport(
            ip=args.device_ip_or_hostname,
        )

    if args.call_type == 'send' and args.payload is None:
        parser.error(
            'argument -p/--payload is required when -c/--call-type is send'
        )

    connector = Connector(
        transport=transport,
    )

    connector.connect()

    if args.call_type == 'learn':
        connector.learn()
    elif args.call_type == 'send':
        connector.send(args.payload)

    connector.disconnect()



import unittest

from hamcrest import assert_that, equal_to
from mock import patch, call, MagicMock

from zmote.connector import HTTPTransport, TCPTransport
from zmote.discoverer_test import _UUID

_TEST_SENDIR_REQUEST = 'sendir,1:1,0,36000,1,1,32,32,64,32,32,64,32,3264'

_TEST_SENDIR_RESPONSE = 'completeir,1:1,0'


class HTTPTransportTest(unittest.TestCase):
    def setUp(self):
        self._subject = HTTPTransport(
            ip='192.168.1.12',
        )

        self._subject._session = MagicMock()
        self._subject._uuid = _UUID

    @patch('zmote.connector.Session')
    def test_connect(self, session):
        session.get.return_value = 'uuid,{0}'.format(_UUID)

        self._subject.connect()

        assert_that(
            session().get.mock_calls[0:1],
            equal_to([
                call('http://{0}/uuid'.format(self._subject._ip), timeout=5),
            ])
        )

    def test_call(self):
        mock_response = MagicMock()
        mock_response.text = _TEST_SENDIR_RESPONSE

        self._subject._session.post.return_value = mock_response

        assert_that(
            self._subject.call(_TEST_SENDIR_REQUEST),
            equal_to(_TEST_SENDIR_RESPONSE)
        )

        assert_that(
            self._subject._session.mock_calls,
            equal_to([
                call.post(
                    url='http://{0}/v2/{1}'.format(self._subject._ip, _UUID),
                    data=_TEST_SENDIR_REQUEST,
                    timeout=5,
                )
            ])
        )

    def test_disconnect(self):
        self._subject.disconnect()

        assert_that(
            self._subject._session,
            equal_to(None)
        )


class TCPTransportTest(unittest.TestCase):
    @patch('zmote.connector.socket')
    def setUp(self, socket):
        socket.AF_INET = 1
        socket.SOCK_STREAM = 2

        self._subject = TCPTransport(
            ip='192.168.1.12',
        )

        self._subject._sock = MagicMock()

        assert_that(
            socket.mock_calls,
            equal_to([
                call.socket(socket.AF_INET, socket.SOCK_STREAM)
            ])
        )

    def test_connect(self):
        self._subject.connect()

        assert_that(
            self._subject._sock.mock_calls,
            equal_to([
                call.connect((self._subject._ip, 4998))
            ])
        )

    def test_call(self):
        self._subject._sock.recv.return_value = _TEST_SENDIR_RESPONSE

        assert_that(
            self._subject.call(_TEST_SENDIR_REQUEST),
            equal_to(_TEST_SENDIR_RESPONSE)
        )

        assert_that(
            self._subject._sock.mock_calls,
            equal_to([
                call.send('sendir,1:1,0,36000,1,1,32,32,64,32,32,64,32,3264'),
                call.recv(1024)
            ])
        )

    def test_disconnect(self):
        self._subject.disconnect()

        assert_that(
            self._subject._sock.mock_calls,
            equal_to([
                call.close()
            ])
        )
